from datetime import datetime, timedelta

import pytest
from rest_framework.exceptions import ValidationError

from django.utils.timezone import make_aware
from program.models import ScheduleConflictError, TimeSlot
from program.services import (
    generate_conflicts,
    generate_timeslots,
    make_conflicts,
    resolve_conflicts,
)
from program.tests.factories import TimeslotFactory
from program.tests.test_schedules import schedule_data
from program.tests.test_schedules import url as schedule_url
from program.typing import ScheduleCreateUpdateData, ScheduleData

pytestmark = pytest.mark.django_db


def create_timeslot(once_schedule) -> None:
    """creates and saves a timeslot instance for a "once" schedule for 2024-05-21 16:00-18:00."""

    TimeslotFactory.create(
        start=make_aware(datetime(2024, 5, 21, 16, 0)),
        end=make_aware(datetime(2024, 5, 21, 18, 0)),
        schedule=once_schedule,
    )


def build_timeslot(once_schedule, start: tuple[int, int], end: tuple[int, int]) -> TimeSlot:
    """builds and returns a timeslot object for a "once" schedule for 2020-05-21 `start`-`end`."""

    start_hour, start_minute = start
    end_hour, end_minute = end

    return TimeslotFactory.build(
        start=make_aware(datetime(2024, 5, 21, start_hour, start_minute)),
        end=make_aware(datetime(2024, 5, 21, end_hour, end_minute)),
        schedule=once_schedule,
    )


def test_generate_conflicts_partially_overlapping_end(once_schedule):
    create_timeslot(once_schedule)  # 16:00 - 18:00
    # Partly overlapping timeslot: start before existing start, end before existing end
    timeslots = [build_timeslot(once_schedule, start=(15, 30), end=(17, 30))]

    conflicts = generate_conflicts(timeslots)

    assert "theirs-end" and "ours-end" in conflicts["projected"][0]["solution_choices"]


def test_generate_conflicts_partially_overlapping_start(once_schedule):
    create_timeslot(once_schedule)  # 16:00 - 18:00
    # Partly overlapping timeslots: start after existing start, end after existing end
    timeslots = [build_timeslot(once_schedule, start=(16, 30), end=(18, 30))]

    conflicts = generate_conflicts(timeslots)

    assert "theirs-start" and "ours-start" in conflicts["projected"][0]["solution_choices"]


def test_generate_conflicts_fully_overlapping_theirs(once_schedule):
    create_timeslot(once_schedule)  # 16:00 - 18:00
    # Fully overlapping timeslot: start before existing start, end after existing end
    timeslots = [build_timeslot(once_schedule, start=(15, 30), end=(18, 30))]

    conflicts = generate_conflicts(timeslots)

    assert (
        "theirs-end"
        and "theirs-start"
        and "theirs-both" in conflicts["projected"][0]["solution_choices"]
    )


def test_generate_conflicts_fully_overlapping_ours(once_schedule):
    create_timeslot(once_schedule)  # 16:00 - 18:00
    # Fully overlapping: start after existing start, end before existing end
    timeslots = [build_timeslot(once_schedule, start=(16, 30), end=(17, 30))]

    conflicts = generate_conflicts(timeslots)

    assert (
        "ours-end"
        and "ours-start"
        and "ours-both" in conflicts["projected"][0]["solution_choices"]
    )


def test_generate_conflicts_overlapping_ours_theirs(once_schedule):
    create_timeslot(once_schedule)  # 16:00 - 18:00
    # Fully overlapping: starts and ends are equal
    timeslots = [build_timeslot(once_schedule, start=(16, 0), end=(18, 0))]

    conflicts = generate_conflicts(timeslots)

    assert "ours" and "theirs" in conflicts["projected"][0]["solution_choices"]


def test_generate_timeslots_once(once_schedule):
    timeslots = generate_timeslots(once_schedule)

    assert len(timeslots) == 1


def test_generate_timeslots_daily(daily_schedule):
    timeslots = generate_timeslots(daily_schedule)

    assert len(timeslots) == 7 or 8


def test_generate_timeslots_weekly(weekly_schedule):
    timeslots = generate_timeslots(weekly_schedule)

    assert len(timeslots) == 4 or 5


def test_generate_timeslots_business_days(business_days_schedule):
    timeslots = generate_timeslots(business_days_schedule)

    assert len(timeslots) == 5 * 4 or 5 * 4 + 1


def test_generate_timeslots_weekends(weekends_schedule):
    timeslots = generate_timeslots(weekends_schedule)

    assert len(timeslots) == 2 * 4 or 2 * 4 + 1  # I’m not sure why, but this changes on weekends


def test_generate_timeslots_monthly(monthly_schedule):
    timeslots = generate_timeslots(monthly_schedule)

    assert len(timeslots) == 1 or 2  # I’m not sure why, but this changes on February


def test_make_conflicts_once_schedule_no_collisions(show, once_schedule):
    data = ScheduleData(
        first_date=once_schedule.first_date.strftime("%Y-%m-%d"),
        end_time=once_schedule.end_time.strftime("%H:%M:%S"),
        rrule_id=once_schedule.rrule_id,
        show_id=show.id,
        start_time=once_schedule.start_time.strftime("%H:%M:%S"),
    )

    conflicts = make_conflicts(data, None, once_schedule.show_id)
    projected = conflicts["projected"][0]

    assert projected["collisions"] == []
    assert projected["error"] is None
    assert projected["solution_choices"] == set()


def test_make_conflicts_weekly_schedule_collisions(admin_api_client, show, weekly_schedule):
    admin_api_client.post(schedule_url(), data=schedule_data(weekly_schedule, show), format="json")

    data = ScheduleData(
        first_date=weekly_schedule.first_date.strftime("%Y-%m-%d"),
        end_time=weekly_schedule.end_time.strftime("%H:%M:%S"),
        rrule_id=weekly_schedule.rrule_id,
        show_id=show.id,
        start_time=weekly_schedule.start_time.strftime("%H:%M:%S"),
    )

    conflicts = make_conflicts(data, None, weekly_schedule.show_id)

    for projected in conflicts["projected"]:
        assert len(projected["collisions"]) == 1
        assert projected["error"] is None
        assert projected["solution_choices"] == {"ours", "theirs"}


def test_make_conflicts_extend_weekly_schedule_collisions(admin_api_client, show, weekly_schedule):
    response = admin_api_client.post(
        schedule_url(), data=schedule_data(weekly_schedule, show), format="json"
    )
    schedule_id = response.json().get("id")

    in_one_year = datetime.now() + timedelta(days=365)
    data = ScheduleData(
        first_date=weekly_schedule.first_date.strftime("%Y-%m-%d"),
        end_time=weekly_schedule.end_time.strftime("%H:%M:%S"),
        last_date=in_one_year.strftime("%Y-%m-%d"),
        rrule_id=weekly_schedule.rrule_id,
        show_id=show.id,
        start_time=weekly_schedule.start_time.strftime("%H:%M:%S"),
    )

    conflicts = make_conflicts(data, schedule_id, weekly_schedule.show_id)

    for projected in conflicts["projected"]:
        assert len(projected["collisions"]) == 1
        assert projected["error"] is None
        assert projected["solution_choices"] == {"ours", "theirs"}


def test_resolve_conflicts_no_same_day_start_and_end(show, weekly_schedule):
    data = ScheduleCreateUpdateData(
        schedule=ScheduleData(
            end_time=weekly_schedule.end_time.strftime("%H:%M:%S"),
            first_date=weekly_schedule.first_date.strftime("%Y-%m-%d"),
            last_date=weekly_schedule.first_date.strftime("%Y-%m-%d"),
            rrule_id=weekly_schedule.rrule_id,
            show_id=show.id,
            start_time=weekly_schedule.start_time.strftime("%H:%M:%S"),
        )
    )

    with pytest.raises(ValidationError, match="Start and end dates must not be the same."):
        resolve_conflicts(data, None, show.pk)


def test_resolve_conflicts_no_start_after_end(show, weekly_schedule):
    data = ScheduleCreateUpdateData(
        schedule=ScheduleData(
            end_time=weekly_schedule.start_time.strftime("%H:%M:%S"),
            first_date=weekly_schedule.first_date.strftime("%Y-%m-%d"),
            last_date=(weekly_schedule.first_date - timedelta(days=1)).strftime("%Y-%m-%d"),
            rrule_id=weekly_schedule.rrule_id,
            show_id=show.id,
            start_time=weekly_schedule.end_time.strftime("%H:%M:%S"),
        )
    )

    with pytest.raises(ValidationError, match="End date mustn't be before start."):
        resolve_conflicts(data, None, show.pk)


def test_resolve_conflicts_one_solution_per_conflict(admin_api_client, show, weekly_schedule):
    admin_api_client.post(schedule_url(), data=schedule_data(weekly_schedule, show), format="json")

    data = ScheduleCreateUpdateData(
        schedule=ScheduleData(
            end_time=weekly_schedule.end_time.strftime("%H:%M:%S"),
            first_date=weekly_schedule.first_date.strftime("%Y-%m-%d"),
            rrule_id=weekly_schedule.rrule_id,
            show_id=show.id,
            start_time=weekly_schedule.start_time.strftime("%H:%M:%S"),
        )
    )

    with pytest.raises(
        ScheduleConflictError, match="Numbers of conflicts and solutions don't match."
    ):
        resolve_conflicts(data, None, show.pk)


def test_resolve_conflicts_unresolved_conflicts(admin_api_client, show, once_schedule):
    admin_api_client.post(schedule_url(), data=schedule_data(once_schedule, show), format="json")

    data = ScheduleCreateUpdateData(
        schedule=ScheduleData(
            end_time=once_schedule.end_time.strftime("%H:%M:%S"),
            first_date=once_schedule.first_date.strftime("%Y-%m-%d"),
            rrule_id=once_schedule.rrule_id,
            show_id=show.id,
            start_time=once_schedule.start_time.strftime("%H:%M:%S"),
        ),
        solutions={"": "ours"},
    )

    with pytest.raises(ScheduleConflictError, match="Not all conflicts have been resolved."):
        resolve_conflicts(data, None, show.pk)


def test_resolve_conflicts_dry_run(admin_api_client, show, weekly_schedule):
    data = ScheduleCreateUpdateData(
        schedule=ScheduleData(
            dryrun=True,
            end_time=weekly_schedule.end_time.strftime("%H:%M:%S"),
            first_date=weekly_schedule.first_date.strftime("%Y-%m-%d"),
            rrule_id=weekly_schedule.rrule_id,
            show_id=show.id,
            start_time=weekly_schedule.start_time.strftime("%H:%M:%S"),
        )
    )

    schedule = resolve_conflicts(data, None, show.pk)

    assert len(schedule["create"]) == 52 or 53
    assert schedule["update"] == []
    assert schedule["delete"] == []


def test_resolve_conflicts_no_errors(show, weekly_schedule):
    data = ScheduleCreateUpdateData(
        schedule=ScheduleData(
            end_time=weekly_schedule.end_time.strftime("%H:%M:%S"),
            first_date=weekly_schedule.first_date.strftime("%Y-%m-%d"),
            rrule_id=weekly_schedule.rrule_id,
            show_id=show.id,
            start_time=weekly_schedule.start_time.strftime("%H:%M:%S"),
        )
    )

    schedule = resolve_conflicts(data, None, show.pk)

    assert "errors" not in schedule
